from pyticketswitch.util import resolve_boolean


class APIException(Exception):
    """General Exception object for representing an error returned by the API.

    Used as a super class for some other Exception objects defined here.
    """

    def __init__(
        self, call, code, description
    ):
        self.call = call
        self.code = code
        self.description = description

    def __str__(self):
        return "API Exception - call={0}, code={1}, description={2}".format(
            self.call, self.code, self.description
        )


class InvalidToken(APIException):
    """Exception indicating that the token passed to the API was invalid.

    Can also be referred to as InvalidId.
    """

    def __init__(
        self, call, description, code=None
    ):
        if code is None:
            code = "N/A"

        super(InvalidToken, self).__init__(
            call=call, code=code, description=description
        )

InvalidId = InvalidToken


class BackendCallFailure(APIException):
    """The API returned that the backend call failed.

    This indicates that there is a problem with supplier's system,
    causing the API call to fail.
    """

    def __init__(
        self, call
    ):
        code = "N/A"
        description = "The call to the backend system failed."

        super(BackendCallFailure, self).__init__(
            call=call, code=code, description=description
        )


class CommsException(Exception):
    """Thrown when there was problem communicating with the API.

    This could be a problem at the HTTP, URL or networking
    level, including a timeout (i.e. the API takes more than
    the timeout settings value to respond).
    """

    def __init__(
        self, underlying_exception, description
    ):
        self.underlying_exception = underlying_exception
        self.description = description

    def __str__(self):
        return self.description


class InvalidResponse(Exception):
    """The response from the API could not be interpreted.

    This is thrown when the data returned by the API cannot
    be parsed, most likely indicates a failure of the API.
    """

    def __init__(
        self, underlying_exception, description
    ):
        self.underlying_exception = underlying_exception
        self.description = description

    def __str__(self):
        return self.description


# TROLLEY ADD ERRORS
class TrolleyAddErrors(Exception):
    """Thrown when an attempt to add an Order to a Trolley fails.

    Used to represent errors returned by the 'trolley_add_order' API
    call. The string representation of the object describes the errors
    in the 'errors' list.

    There is more detail about errors when adding orders to the trolley
    in the XML API documentation.

    Attributes:
        errors (list): List of Exception objects representing errors
            returned when adding an Order to a Trolley.
    """

    def __init__(self, errors):
        self.errors = errors

    def __str__(self):
        return ', '.join([str(type(error)) for error in self.errors])


class TrolleyAddBadBundle(Exception):
    """Reached the maximum number of orders for the supplier's system.

    Attributes:
        max_size (int): The maximum number of orders this supplier allows
            per transaction.
    """

    def __init__(
        self, max_size
    ):
        self.max_size = max_size

    def __str__(self):
        return "Trolley can only contain {0} orders from this supplier".format(
            self.max_size
        )


class TrolleyAddBadCombo(Exception):
    """Order could not be added to the trolley because the mix of suppliers
    is not allowed.

    Attributes:
        system (string): Code of existing supplier.
        system_desc (string): Name of existing supplier.
    """

    def __init__(self, system, system_desc):
        self.system = system
        self.system_desc = system_desc

    def __str__(self):
        return (
            "Supplier is not compatible with {1} ({0}), which is " +
            "currently in the trolley"
        ).format(
            self.system, self.system_desc
        )


class TrolleyAddBadCardTypes(Exception):
    """Order could not be added because there are no compatible card types
    between suppliers.
    """

    def __str__(self):
        return "No compatible card types between suppliers"


class TrolleyAddBadCountries(Exception):
    """Order could not be added due to incompatible countries allowed by the
    suppliers."""

    def __str__(self):
        return "Incompatible allowed countries"


class TrolleyAddBadCurrencyMix(Exception):
    """Order could not be added as there is an existing order for this supplier
    in the trolley with a different currency."""

    def __init__(
        self, system, system_desc, currency, currency_name
    ):
        self.system = system
        self.system_desc = system_desc
        self.currency = currency
        self.currency_name = currency_name

    def __str__(self):
        return "Currency conflicts with existing order on {1} ({0})".format(
            self.system, self.system_desc
        )


class TrolleyAddBadDepart(Exception):
    """Order could not be added due to incompatible departure dates."""

    def __str__(self):
        return "Incompatible departure dates"


class TrolleyAddBadSend(Exception):
    """Order could not be added due to incompatible despatch methods."""

    def __str__(self):
        return "Incompatible despatch methods"
# END TROLLEY ADD ERRORS


# TROLLEY ERRORS
class TrolleyPurchased(APIException):
    """Thrown when the operation attempted is not possible as the trolley has
    already been purchased."""
    pass


class TrolleyReserved(APIException):
    """Thrown when the operation attempted is not possible as the trolley has
    already been reserved."""
    pass


class NoAvailability(APIException):
    """Thrown when the operation failed due to there no longer being
    availability for the seats."""

    def __init__(
        self, call
    ):
        self.call = call
        code = "N/A"
        description = "The call returned no availability"

        super(NoAvailability, self).__init__(
            call=call, code=code, description=description
        )
# END TROLLEY ERRORS


# PURCHASE ERRORS
class PurchaseValidationError(APIException):
    """Super class for purchase validation errors."""
    pass


class InvalidCountryForDespatch(PurchaseValidationError):
    """The chosen despatch method does not allow the given country code
    in the customer data (XML API code 1306)."""
    pass


class InvalidEmailAddress(PurchaseValidationError):
    """The supplied email address failed RFC822 syntax checking
    (XML API code 1307)."""
    pass


class IncompleteCustomerDetails(PurchaseValidationError):
    """The customer details supplied are incomplete (XML API code 1308)."""
    pass


class NoCardNumberProvided(PurchaseValidationError):
    """No card_number element was present in the payment card data
    (XML API code 1309)."""
    pass


class UnknownCardType(PurchaseValidationError):
    """The payment card type is not known from the supplied card number
    (XML API code 1310)."""
    pass


class InvalidCardType(PurchaseValidationError):
    """The payment card type is not one of those accepted for this
    transaction (XML API code 1311)."""
    pass


class InvalidCardNumber(PurchaseValidationError):
    """The card number given is not valid for cards of that type
    (XML API code 1312)."""
    pass


class NoExpiryDateProvided(PurchaseValidationError):
    """No "expiry_date" element was present in the payment card data
    (XML API code 1313)."""
    pass


class InvalidExpiryDate(PurchaseValidationError):
    """The expiry date given is not valid (XML API code 1314)."""
    pass


class NoCV2Provided(PurchaseValidationError):
    """No 'cv_two' was present in the payment card data (XML API code 1315)."""
    pass


class InvalidCV2(PurchaseValidationError):
    """The CV2 value given is not valid (XML API code 1316)."""
    pass


class NoIssueNumberProvided(PurchaseValidationError):
    """No 'issue_number' has been supplied when required (XML API code 1317).
    """
    pass


class InvalidIssueNumber(PurchaseValidationError):
    """The issue number given is not valid (XML API code 1319)."""
    pass


class IncompleteAltBillingAddress(PurchaseValidationError):
    """The billing address details supplied are incomplete
    (XML API code 1321)."""
    pass


class NoStartDateProvided(PurchaseValidationError):
    """No 'start_date' element has been supplied when required
    (XML API code 1322)."""
    pass


class InvalidStartDate(PurchaseValidationError):
    """The start date given is not valid (XML API code 1323)."""
    pass


class ReservationExpired(APIException):
    """Operation failed because the reservation has expired
    (XML API code 1404)."""
    pass


class PurchaseException(APIException):
    """General Exception class for purchase errors.

    This class is sub classed by several Exception classes.
    """
    def __init__(
        self, call, description, code=None, failed_cv_two=None,
        failed_avs=None, failed_3d_secure=None
    ):
        self._failed_cv_two = failed_cv_two
        self._failed_avs = failed_avs
        self._failed_3d_secure = failed_3d_secure

        super(PurchaseException, self).__init__(
            call=call, code=code, description=description
        )

    @property
    def failed_cv_two(self):
        """Boolean to indicate if the purchase attempt was marked with a
        CV2 failure."""
        return resolve_boolean(self._failed_cv_two)

    @property
    def failed_avs(self):
        """Boolean to indicate if the purchase attempt was marked with a
        AVS failure."""
        return resolve_boolean(self._failed_avs)

    @property
    def failed_3d_secure(self):
        """Boolean to indicate if the purchase attempt was marked with a
        3D secure failure."""
        return resolve_boolean(self._failed_3d_secure)


class FraudTriggered(PurchaseException):
    """The combination of customer and payment card details have matched a
    system fraud trigger and the purchase has not be allowed.

    (XML API purchase_fail_code 1).
    """
    pass


class CardAuthorisationFailed(PurchaseException):
    """The payment card provided could not be authorised for the value of
    the transaction or the authorisation timed out.

    (XML API purchase_fail_code 2, 3).
    """
    pass


class ReservationAlreadyPurchased(PurchaseException):
    """The reservation has already been purchased (XML API
    purchase_fail_code 4)."""
    pass


class PurchasePreviouslyAttempted(PurchaseException):
    """The reservation purchase has already been attempted and failed
    (XML API purchase_fail_code 5)."""
    pass


class PurchaseRefused(PurchaseException):
    """The reservation purchase has been refused by the backend
    (XML API purchase_fail_code 6)."""
    pass


# 7 - an unspecified systems error has occurred during purchase
class PurchaseFailed(PurchaseException):
    """An unspecified systems error has occurred during purchase
    (XML API purchase_fail_code 7)."""
    pass
# END PURCHASE ERRORS



from operator import itemgetter, attrgetter
import datetime
from copy import deepcopy

from pyticketswitch.util import (
    resolve_boolean, to_int_or_none, yyyymmdd_to_date,
    dates_in_range, hhmmss_to_time, date_to_yyyymmdd_or_none
)
from base import InterfaceObject, CostRangeMixin
from pyticketswitch.api_exceptions import InvalidId
from pyticketswitch import settings
import core as core_objs
import performance as perf_objs
import availability as avail_objs


class Category(object):
    """Object representing a TSW event category.

    In the XML API documentation this is also referred to as the
    event class
    """

    def __init__(
        self,
        core_class=None,
        core_subclass=None
    ):
        self._core_class = core_class
        if not core_class:
            self._core_subclass = core_subclass
        self._sub_categories = None

    @property
    def code(self):
        if self._core_class:
            return self._core_class.class_code
        else:
            return self._core_subclass.subclass_code

    @property
    def description(self):
        if self._core_class:
            return self._core_class.class_desc
        else:
            return self._core_subclass.subclass_desc

    @property
    def is_main(self):
        """Flag to indicate if this is the primary category for the event."""
        if self._core_class:
            return resolve_boolean(self._core_class.is_main_class)
        else:
            return resolve_boolean(self._core_subclass.is_main_subclass)

    @property
    def search_key(self):
        """String representation for use in the event search"""
        if self._core_class:
            return self._core_class.search_key
        else:
            return self._core_subclass.search_key

    @property
    def slug_search_key(self):
        """Slug of the search key"""
        return self.search_key.replace('/', '_')

    @staticmethod
    def deslug_search_key(slug_search_key):
        """Returns the search key from a slugged search key.

        Args:
            slug_search_key (string): A slug_search_key

        Returns:
            string: The search_key
        """
        return slug_search_key.replace('_', '/')

    @property
    def sub_categories(self):
        """List of sub categories"""
        if self._core_class and self._sub_categories is None:

            self._sub_categories = []

            for sub in self._core_class.subclasses:

                self._sub_categories.append(
                    Category(core_subclass=sub)
                )

        return self._sub_categories


class CustomField(object):
    """Represents a TSW event custom field."""

    def __init__(
        self,
        core_custom_field,
    ):
        self._core_custom_field = core_custom_field

    @property
    def description(self):
        if self._core_custom_field.custom_field_label:
            return self._core_custom_field.custom_field_label
        return None

    @property
    def code(self):
        return self._core_custom_field.custom_field_name

    @property
    def key(self):
        return self.code

    @property
    def values(self):
        return self._core_custom_field.custom_field_data.split(', ')


class CustomFilter(object):
    """Represents a TSW event custom filter."""

    def __init__(
        self,
        core_custom_filter,
    ):
        self._core_custom_filter = core_custom_filter

    @property
    def description(self):
        return self._core_custom_filter.custom_filter_desc

    @property
    def key(self):
        return self._core_custom_filter.custom_filter_key


class StructuredContentItem(object):
    """Represents an item of Structured Content.

    For example, Events may have an associated 'Overview' Structured
    Content Item, which would have a key of 'overview'. The name would
    then be the translated name of the item (so 'Overview' in English)
    and the value would be the translated value of this item for the Event.
    """

    def __init__(
        self,
        core_structured_info_item,
    ):
        self._core_structured_info_item = core_structured_info_item

    @property
    def key(self):
        return self._core_structured_info_item.key

    @property
    def name(self):
        return self._core_structured_info_item.name

    @property
    def value(self):
        return self._core_structured_info_item.value


class Event(InterfaceObject, CostRangeMixin):
    """Object that represents a TSW event.

    Operations relating to TSW events are accessible through here,
    e.g. retrieving event information, listing performances, etc.

    The Event Id (event_id) is the only required argument apart from
    configuration settings, which are explained in the Core object
    constructor. The other arguments are for internal use.

    Inherits from the CostRangeMixin class, so these functions are
    available.

    Args:
        event_id (string): The Id of the Event object.
        settings (kwargs): See Core constructor.
    """

    _attr_request_map = {
        'event_info': 'extra_info',
        'venue_info': 'extra_info',
        'venue_addr': 'extra_info',
        'source_t_and_c': 'source_info',
        'critic_review_percent': 'extra_info',
        'user_review_percent': 'extra_info',
        'reviews': 'reviews',
        'cost_range': 'cost_range',
        'video_iframe': 'video_iframe',
        'custom_fields': 'custom_fields',
        'event_medias': 'media',
        'structured_info': 'extra_info_only',
        'event_quantity_options': 'extra_info_only',
        'avail_details': 'avail_details',
    }

    def __init__(
        self,
        event_id,
        core_event=None,
        requested_data=None,
        **settings
    ):

        self.event_id = event_id
        self._core_event = core_event
        self._performance_calendar = {}
        self._performances = None
        self._need_departure_date = None
        self._categories = None
        self._critic_reviews = None
        self._user_reviews = None
        self._video = None
        self._months = None
        self._custom_fields = None
        self._custom_filters = None
        self._perfs_have_usage_date = None
        self._perfs_have_required_info = None
        self._has_single_false_perf = None
        self._structured_content = None
        self._valid_ticket_quantities = None
        self._avail_details = None

        if not requested_data:
            self._requested_data = {}
        else:
            self._requested_data = deepcopy(requested_data)

        super(Event, self).__init__(**settings)

    def __getstate__(self):
        """This method determines what data will be pickled, and therefore,
        what data will be stored in the event cache.
        Any attributes that should not be stored in the cache should be
        set to None in the returned dictionary.
        """

        d = super(Event, self).__getstate__()

        # Don't cache avail details
        d['_avail_details'] = None
        # Remove avail details from requested data to make sure it is
        # requested again next time
        if self._attr_request_map['avail_details'] in self._requested_data:
            del self._requested_data[self._attr_request_map['avail_details']]

        return d

    def _get_cache_key(self):
        return self.event_id

    def _get_core_event_attr(self, attr_name):

        if not self._core_event:
            try:
                self.get_details()
            except InvalidId:
                return None

        attr_value = getattr(self._core_event, attr_name, None)

        if not attr_value:
            if attr_name in self._attr_request_map:
                request_val = self._attr_request_map[attr_name]

                if request_val not in self._requested_data:
                    self.get_details()

                    attr_value = getattr(self._core_event, attr_name, None)

        return attr_value

    # Overridden function for CostRangeMixin
    def _get_core_cost_range(self):
        return self._get_core_event_attr('cost_range')

    @property
    def description(self):
        return self._get_core_event_attr('event_desc')

    @property
    def information(self):
        return self._get_core_event_attr('event_info')

    @property
    def venue_desc(self):
        return self._get_core_event_attr('venue_desc')

    @property
    def venue_info(self):
        return self._get_core_event_attr('venue_info')

    @property
    def venue_addr(self):
        return self._get_core_event_attr('venue_addr')

    @property
    def supplier_code(self):
        return self._get_core_event_attr('source_code')

    @property
    def supplier_desc(self):
        return self._get_core_event_attr('source_desc')

    @property
    def supplier_t_and_c(self):
        return self._get_core_event_attr('source_t_and_c')

    @property
    def city_code(self):
        return self._get_core_event_attr('city_code')

    @property
    def city_desc(self):
        return self._get_core_event_attr('city_desc')

    @property
    def country_code(self):
        return self._get_core_event_attr('country_code')

    @property
    def country_desc(self):
        return self._get_core_event_attr('country_desc')

    @property
    def is_seated(self):
        return resolve_boolean(
            self._get_core_event_attr('is_seated')
        )

    @property
    def has_no_perfs(self):
        return resolve_boolean(
            self._get_core_event_attr('has_no_perfs')
        )

    @property
    def upsell_list(self):
        return self._get_core_event_attr('event_upsell_list')

    @property
    def show_perf_time(self):
        return resolve_boolean(
            self._get_core_event_attr('show_perf_time')
        )

    @property
    def need_performance(self):
        return resolve_boolean(
            self._get_core_event_attr('need_performance')
        )

    @property
    def need_duration(self):
        return resolve_boolean(
            self._get_core_event_attr('need_duration')
        )

    @property
    def valid_ticket_quantities(self):
        """Returns a list of the valid ticket quantities."""

        if self._valid_ticket_quantities is None:

            self._valid_ticket_quantities = []
            # Call extra_info to make sure we get latest data
            # (Important for redeem users for example, since we don't want
            # to get old cached data)
            self.get_details()

            if self.cached_valid_ticket_quantities:
                self._valid_ticket_quantities = (
                    self.cached_valid_ticket_quantities
                )

            else:
                quantity_options_dict = self._get_core_event_attr(
                    'event_quantity_options'
                )
                if quantity_options_dict:
                    valid_list = quantity_options_dict.get(
                        'valid_quantity', []
                    )
                    self._valid_ticket_quantities = [
                        int(x) for x in valid_list
                    ]

        return self._valid_ticket_quantities

    @property
    def latitude(self):
        geo_data = self._get_core_event_attr('geo_data')

        if geo_data:
            return geo_data.latitude

        else:
            return None

    @property
    def longitude(self):
        geo_data = self._get_core_event_attr('geo_data')

        if geo_data:
            return geo_data.longitude

        else:
            return None

    @property
    def critic_review_percent(self):
        rev_per = self._get_core_event_attr('critic_review_percent')

        if rev_per:
            if rev_per == '0':

                rev_per = None
            else:
                rev_per = rev_per + '%'

        return rev_per

    @property
    def user_review_percent(self):
        rev_per = self._get_core_event_attr('user_review_percent')

        if rev_per:
            if rev_per == '0':

                rev_per = None
            else:
                rev_per = rev_per + '%'

        return rev_per

    def _build_reviews(self):
        self._critic_reviews = []
        self._user_reviews = []
        for r in self._get_core_event_attr('reviews'):

            is_user_bool = resolve_boolean(r.is_user_review)

            review = Review(
                core_review=r, is_user_review=is_user_bool
            )

            if is_user_bool:
                self._user_reviews.append(review)
            else:
                self._critic_reviews.append(review)

    @property
    def critic_reviews(self):
        """List of Review objects."""
        if self._critic_reviews is None:
            self._build_reviews()

        return self._critic_reviews

    @property
    def user_reviews(self):
        """List of Review objects."""
        if self._user_reviews is None:
            self._build_reviews()

        return self._user_reviews

    @property
    def categories(self):
        """List of Category objects."""

        if self._categories is None:

            self._categories = []

            for c in self._get_core_event_attr('classes'):
                #  self._categories[c.class_code] = c.class_desc

                self._categories.append(Category(core_class=c))

        return self._categories

    @property
    def video(self):
        """Video object for this Event."""
        if self._video is None:

            core_video = self._get_core_event_attr('video_iframe')

            if core_video:

                self._video = Video(
                    core_video_iframe=core_video
                )

        return self._video

    @property
    def start_date(self):
        start_date = None
        date_range_start = self._get_core_event_attr('date_range_start')

        if date_range_start:
            start_date = yyyymmdd_to_date(
                date_range_start['date_yyyymmdd']
            )

        return start_date

    @property
    def end_date(self):
        end_date = None
        date_range_end = self._get_core_event_attr('date_range_end')

        if date_range_end:
            end_date = yyyymmdd_to_date(
                date_range_end['date_yyyymmdd']
            )

        return end_date

    @property
    def custom_fields(self):
        """Returns list of CustomField objects"""
        if self._custom_fields is None:

            self._custom_fields = []

            for c in self._get_core_event_attr('custom_fields'):
                self._custom_fields.append(
                    CustomField(core_custom_field=c)
                )

        return self._custom_fields

    @property
    def custom_filters(self):
        """Returns list of CustomFilter objects"""
        if self._custom_filters is None:

            self._custom_filters = []

            for c in self._get_core_event_attr('custom_filters'):
                self._custom_filters.append(
                    CustomFilter(core_custom_filter=c)
                )

        return self._custom_filters

    def _get_image(self, name, request_all_images=True):

        make_request = True

        if (
            self._attr_request_map['event_medias'] in
            self._requested_data
        ):

            for req in self._requested_data[
                self._attr_request_map['event_medias']
            ].keys():
                if req in name or req == name:
                    make_request = False
                    break

        if make_request:
            if request_all_images:
                request_media = None
            else:
                request_media = [name]

            self.get_details(request_media=request_media)

        event_medias = self._get_core_event_attr('event_medias')

        if event_medias:
            for i in event_medias:
                if i.name == name:
                    return i.secure_complete_url

        return None

    def get_large_image(self):
        return self._get_image(settings.LARGE_IMAGE)

    def get_large_images(self):

        images = []

        i = 0

        for name in settings.LARGE_IMAGE_LIST:

            url = self._get_image(name)

            if url:
                images.append(
                    {'url': url, 'index': i}
                )
                i = i + 1

        return images

    def get_medium_image(self):
        return self._get_image(settings.MEDIUM_IMAGE)

    def get_small_image(self):
        return self._get_image(settings.SMALL_IMAGE)

    def get_marquee_image(self):
        return self._get_image(settings.MARQUEE_IMAGE)

    def get_seating_plan(self):
        return self._get_image(settings.SEATING_IMAGE)

    def get_supplier_image(self):
        return self._get_image(settings.SUPPLIER_IMAGE)

    def get_details(
        self, request_media=None, source_info=True,
        request_reviews=True, request_avail_details=None,
        mime_text_type='html',
    ):
        """Retrieves data for the current Event.

        This is called internally to get information about
        the current Event, but can be called explicitly
        if required (could reduce the number of API calls
        in some circumstances).
        """

        if request_media is None:
            request_media = settings.REQUEST_MEDIA

        crypto_block = self.get_crypto_block(
            method_name='event_search',
        )

        detailed_event = None
        extra_info_called = False

        if crypto_block:

            detailed_event = self.get_core_api().extra_info(
                crypto_block=crypto_block,
                upfront_data_token=self.settings['upfront_data_token'],
                event_token=self.event_id,
                request_media=request_media, source_info=source_info,
                request_avail_details=request_avail_details,
                mime_text_type=mime_text_type,
            )

            request_reviews = True
            extra_info_called = True

        else:
            core = core_objs.Core(
                **self._internal_settings()
            )
            events = core.search_events(
                event_id_list=[self.event_id], request_media=request_media,
                request_source_info=source_info, request_extra_info=True,
                request_video_iframe=True, request_cost_range=True,
                request_custom_fields=True, request_reviews=request_reviews,
                request_avail_details=request_avail_details,
                mime_text_type=mime_text_type,
            )
            if events:
                detailed_event = events[0]._core_event

            else:
                raise InvalidId(
                    call="event_search",
                    description="Event does not exist"
                )

        if self._core_event is None:
            self._core_event = detailed_event

        else:
            self._core_event.add_extra_info(detailed_event)

        self._requested_data['extra_info'] = True
        self._requested_data['video_iframe'] = True
        self._requested_data['cost_range'] = True
        self._requested_data['custom_fields'] = True

        if source_info:
            self._requested_data['source_info'] = True

        if request_media:
            if 'media' not in self._requested_data:
                self._requested_data['media'] = {}
            for m in request_media:
                self._requested_data['media'][m] = True

        if request_reviews:
            self._requested_data['reviews'] = True

        if request_avail_details:
            self._requested_data['avail_details'] = True

        if extra_info_called:
            self._requested_data['extra_info_only'] = True

    @property
    def performance_calendar(self):
        """Dictionary of Performances for this Event by date.

        Performance objects are returned as a dictionary,
        organised by Year -> Month -> Day, with a list of
        all the performances that day.
        """
        if not self._performance_calendar:

            for performance in self.performances:

                if performance.date:
                    year = performance.date.year
                    month = performance.date.month
                    day = performance.date.day

                    if performance.time:
                        time_desc = '.'.join((str(
                            int(performance.time.strftime('%I'))),
                            performance.time.strftime('%M%p')
                        ))
                    else:
                        time_desc = self.settings['no_time_descr']

                    if year not in self._performance_calendar:
                        self._performance_calendar[year] = {}

                    if month not in self._performance_calendar[year]:
                        self._performance_calendar[year][month] = {}

                    if (
                        day not in
                        self._performance_calendar[year][month]
                    ):
                        self._performance_calendar[year][month][day] = []

                    if (
                        time_desc not in
                        self._performance_calendar[year][month][day]
                    ):
                        self._performance_calendar[year][month][day].append(
                            {'time': time_desc, 'performance': performance}
                        )

        return self._performance_calendar

    def _get_search_crypto(self):

        crypto_block = self.get_crypto_block(
            method_name='event_search',
        )

        if not crypto_block:

            core = core_objs.Core(
                **self._internal_settings()
            )
            events = core.search_events(
                event_id_list=[self.event_id]
            )
            if events:
                self._core_event = events[0]._core_event
            else:
                raise InvalidId(
                    call="event_search",
                    description="Event does not exist"
                )

            crypto_block = self.get_crypto_block(
                method_name='event_search',
            )

        return crypto_block

    @property
    def months(self):
        """List of months that have performances.

        Returns a list of dictionaries with 'month' and 'year'
        keys. Each dictionary represents a month that contains
        Performances of this Event.
        """
        if self._months is None:
            return self.get_valid_months()
        else:
            return self._months

    @months.setter
    def months(self, value):
        self._months = value

    def get_valid_months(self):
        """Retrieves the list of months that contain performances.

        This method is called internally by the 'months' property,
        but can be called explicitly if required.
        """

        crypto_block = self._get_search_crypto()

        resp_dict = self.get_core_api().month_options(
            crypto_block=crypto_block,
            upfront_data_token=self.settings['upfront_data_token'],
            event_token=self.event_id
        )

        months = []

        for m in resp_dict['months']:
            months.append({
                'year': int(m.year_number),
                'month': int(m.month_number),
            })

        self.months = sorted(
            months, key=itemgetter('year', 'month')
        )

        return self.months

    @property
    def performances(self):
        """List of Performance objects for this Event."""
        if self._performances is None:
            return self.get_performances()
        else:
            return self._performances

    @performances.setter
    def performances(self, value):
        self._performances = value

    @property
    def need_departure_date(self):
        """Flag to indicate that a departure date is required.

        Value can be True/False (a departure date is/isn't required)
        or None (the data could not be retrieved/unknown).
        See the API documentation for more information about the
        'need_departure_date' flag.
        """

        if self._need_departure_date is None:
            need_departure_date = self._get_core_event_attr(
                'need_departure_date'
            )

            if need_departure_date:

                self._need_departure_date = resolve_boolean(
                    need_departure_date
                )

            else:
                self.get_performances()

        if self._need_departure_date:
            return self._need_departure_date
        else:
            return None

    @need_departure_date.setter
    def need_departure_date(self, value):
        if value is None:
            self._need_departure_date = False
        else:
            self._need_departure_date = value

    @property
    def perfs_have_usage_date(self):
        """Flag to indicate if the Performances are using usage dates.

        True if the Performances for this Event are using a usage date
        (see the API documentation for more information).
        """
        if self._perfs_have_usage_date is None:
            self.get_performances()

        return self._perfs_have_usage_date

    @perfs_have_usage_date.setter
    def perfs_have_usage_date(self, value):
        self._perfs_have_usage_date = value

    @property
    def perfs_have_required_info(self):
        """Flag to indicate if any of the Performances have
        'required_info' text.

        Value will be True if any of the Performances for this
        Event has required_info text that must be displayed.
        """
        if self.perfs_have_usage_date or self.has_single_false_perf:
            return None

        elif self._perfs_have_required_info is None:
            self.get_performances()

        return self._perfs_have_required_info

    @perfs_have_required_info.setter
    def perfs_have_required_info(self, value):
        self._perfs_have_required_info = value

    @property
    def has_single_false_perf(self):
        """Flag to indicate that there is only one Performance.

        The value will be True if this Event only has a single
        Performance that has been created by this library to
        standardise the booking process. If this flag is True,
        there is no need to display the Performances for this
        Event.
        """
        if self.need_performance or self.need_departure_date:
            return False

        elif self._has_single_false_perf is None:
            self.get_performances()

        return self._has_single_false_perf

    @has_single_false_perf.setter
    def has_single_false_perf(self, value):
        self._has_single_false_perf = value

    @property
    def show_performances(self):
        """Flag to indicate whether to display individual Performances.

        Value will be False if there is no need to show any more
        information about a Performance other than the date.
        """
        if self.has_single_false_perf:
            return False
        elif not self.show_perf_time:
            if self.perfs_have_usage_date:
                return False
            elif not self.perfs_have_required_info:
                return False

        return True

    def get_performances(self, earliest_date=None, latest_date=None, **kwargs):
        """Retrieves the Performances for this Event.

        Returns the list of Performances for the Event, called internally
        by several other methods, it calls the 'date_time_options' API method.
        The 'performances' property should be used to get this information
        unless a specific date range is required.

        Args:
            earliest_date (datetime.date): restrict the list of
                Performances to be later than this date.
            latest_date (datetime.date): restrict the list of
                Performances to be earlier than this date.

        Returns:
            list: List of Performance objects
        """
        crypto_block = self._get_search_crypto()

        resp_dict = self.get_core_api().date_time_options(
            crypto_block=crypto_block,
            upfront_data_token=self.settings['upfront_data_token'],
            event_token=self.event_id,
            earliest_date=date_to_yyyymmdd_or_none(earliest_date),
            latest_date=date_to_yyyymmdd_or_none(latest_date),
            request_cost_range=True,
            **kwargs
        )

        performances = []

        self.need_departure_date = resolve_boolean(
            resp_dict['need_departure_date']
        )

        if 'using_perf_list' in resp_dict:

            self.perfs_have_required_info = resolve_boolean(
                resp_dict['has_perf_names']
            )

            self.perfs_have_usage_date = False
            self.has_single_false_perf = False

            for p in resp_dict['using_perf_list']['performances']:

                if self.need_departure_date:
                    departure_date = p.date
                else:
                    departure_date = None

                performances.append(
                    perf_objs.Performance.from_event_and_perf_token(
                        event=self,
                        perf_token=p.perf_token,
                        core_performance=p,
                        departure_date=departure_date,
                        **self._internal_settings()
                    )
                )

        elif 'using_usage_date' in resp_dict:

            self.perfs_have_usage_date = True
            self.has_single_false_perf = False
            self.perfs_have_required_info = False

            performances = self._build_performances_from_usage(
                usage_date_dict=resp_dict['using_usage_date'],
                need_departure_date=self.need_departure_date,
                latest_date=latest_date
            )

        else:

            self.perfs_have_usage_date = False
            self.has_single_false_perf = True
            self.perfs_have_required_info = False

            performances.append(perf_objs.Performance.from_event_only(
                event=self,
                **self._internal_settings()
            ))

        self.performances = performances

        self._set_crypto_for_object(
            crypto_block=resp_dict['crypto_block'],
            method_name='date_time_options',
            interface_object=self
        )

        return performances

    def _build_performances_from_usage(
        self, usage_date_dict, need_departure_date, latest_date
    ):

        inv_dates = []

        if 'invalid_range' in usage_date_dict:
            usage_inv_range = usage_date_dict['invalid_range']
            if type(usage_inv_range) == list:

                for inv in usage_inv_range:
                    inv_first_date = yyyymmdd_to_date(
                        inv['first_invalid_date_yyyymmdd']
                    )

                    inv_last_date = yyyymmdd_to_date(
                        inv['last_invalid_date_yyyymmdd']
                    )

                    inv_dates.extend(dates_in_range(
                        inv_first_date, inv_last_date
                    ))

            else:
                inv_first_date = yyyymmdd_to_date(
                    usage_inv_range['first_invalid_date_yyyymmdd']
                )

                inv_last_date = yyyymmdd_to_date(
                    usage_inv_range['last_invalid_date_yyyymmdd']
                )

                inv_dates.extend(dates_in_range(inv_first_date, inv_last_date))

        inv_weekdays = []

        if 'invalid_weekday' in usage_date_dict:
            usage_inv_weekdays = usage_date_dict['invalid_weekday']

            if type(usage_inv_weekdays) == list:

                for inv_weekday in usage_inv_weekdays:
                    inv_weekdays.append(inv_weekday['weekday_number'])

            else:
                inv_weekdays.append(int(usage_inv_weekdays['weekday_number']))

        performances = []

        first_date = yyyymmdd_to_date(
            usage_date_dict['first_valid_date_yyyymmdd']
        )

        last_date = yyyymmdd_to_date(
            usage_date_dict['last_valid_date_yyyymmdd']
        )

        if latest_date:

            if latest_date < last_date:
                last_date = latest_date

        delta = last_date - first_date

        for i in range(delta.days + 1):
            perf_date = first_date + datetime.timedelta(days=i)

            if (
                perf_date.isoweekday() not in inv_weekdays and
                perf_date not in inv_dates
            ):

                if need_departure_date:
                    departure_date = perf_date
                else:
                    departure_date = None

                performances.append(
                    perf_objs.Performance.from_event_and_usage_date(
                        event=self,
                        usage_date=perf_date,
                        departure_date=departure_date,
                        **self._internal_settings()
                    )
                )

        return performances

    @property
    def structured_content(self):
        """Dictionary of StructuredContentItem objects for this Event.

        The dictionary keys will be the key of the StructuredContentItem.
        """
        if self._structured_content is None:

            self._structured_content = {}

            si_dict = self._get_core_event_attr('structured_info')

            for key, item in si_dict.items():
                self._structured_content[key] = StructuredContentItem(
                    core_structured_info_item=item,
                )

        return self._structured_content

    def _build_avail_details(self):
        """Builds a list of AvailDetail objects
        """

        self._avail_details = []

        ad_attr = self._get_core_event_attr('avail_details')
        if ad_attr:
            for tt in ad_attr.get('ticket_types', []):
                for pb in tt.get('price_bands', []):
                    for ad in pb.get('avail_details', []):
                        self._avail_details.append(
                            avail_objs.AvailDetail(
                                core_avail_detail=ad,
                                ticket_type_code=tt['ticket_type_code'],
                                ticket_type_desc=tt['ticket_type_desc'],
                                price_band_code=pb['price_band_code'],
                                price_band_desc=pb['price_band_desc'],
                            )
                        )

    @property
    def avail_details(self):
        """Get detailed availability and pricing information for this event.
        """

        if self._avail_details is None:

            if (
                self._attr_request_map['avail_details'] not in
                self._requested_data
            ):
                self.get_details(request_avail_details=True)

            self._build_avail_details()

        return self._avail_details

    @property
    def avail_details_by_price(self):
        """Avail details sorted by combined price
        """

        return sorted(
            self.avail_details,
            key=attrgetter('price_combined_float')
        )

    @property
    def avail_details_by_cheapest_ticket_type(self):
        """Avail details sorted by cheapest ticket type then price. Avail
        details with the same ticket type and pricing will be combined to
        reduce duplicates.
        """
        order_dict = {}

        i = 0
        for ad in self.avail_details_by_price:
            if ad.ticket_type_desc not in order_dict:
                order_dict[ad.ticket_type_desc] = i
                i += 1

        ad_sorted = sorted(
            self.avail_details, key=lambda x: (
                order_dict[x.ticket_type_desc],
                x.price_combined_float
            )
        )

        # Combine rows for which the ticket type and pricing info is the same
        ad_final = []
        for ad in ad_sorted:
            if ad_final and ad.is_same_ticket_and_price(ad_final[-1]):
                ad_final[-1] = ad.combine(ad_final[-1])
            else:
                ad_final.append(ad)

        return ad_final


class Video(object):
    """Object representing a TSW video iframe element."""

    def __init__(
        self,
        core_video_iframe,
    ):
        self._core_video_iframe = core_video_iframe

    @property
    def height(self):
        return to_int_or_none(
            self._core_video_iframe.video_iframe_height
        )

    @property
    def width(self):
        return to_int_or_none(
            self._core_video_iframe.video_iframe_width
        )

    @property
    def host(self):
        return self._core_video_iframe.video_iframe_host

    @property
    def path(self):
        return self._core_video_iframe.video_iframe_path

    @property
    def supports_https(self):
        """Boolean flag indicating HTTPS support."""
        return resolve_boolean(
            self._core_video_iframe.video_iframe_supports_https
        )

    @property
    def secure_url(self):

        if self._core_video_iframe.video_iframe_url_when_secure:
            return self._core_video_iframe.video_iframe_url_when_secure

        else:
            "https://{host}{path}".format(
                host=self.host, path=self.path
            )

    @property
    def insecure_url(self):

        if self._core_video_iframe.video_iframe_url_when_insecure:
            return self._core_video_iframe.video_iframe_url_when_insecure

        else:
            "http://{host}{path}".format(
                host=self.host, path=self.path
            )


class Review(object):
    """Object representing a TSW event review.

    Attributes:
        is_user_review (boolean): Indicates if it is a user review,
            otherwise it is a critic review.
    """

    def __init__(
        self,
        core_review=None,
        is_user_review=None
    ):
        self._core_review = core_review
        self.is_user_review = is_user_review

        if core_review is not None:
            self._set_datetime(
                date_yyyymmdd=core_review.review_date_yyyymmdd,
                time_hhmmss=core_review.review_time_hhmmss
            )

    def _set_datetime(self, date_yyyymmdd, time_hhmmss=None):

        if date_yyyymmdd:
            self.date = yyyymmdd_to_date(date_yyyymmdd)

        else:
            self.date = None

        if time_hhmmss:
            self.time = hhmmss_to_time(time_hhmmss)

        else:
            self.time = None

    @property
    def title(self):
        return self._core_review.review_title

    @property
    def body(self):
        return self._core_review.review_body

    @property
    def author(self):
        return self._core_review.review_author

    @property
    def date_desc(self):
        return self._core_review.review_date_desc

    @property
    def time_desc(self):
        return self._core_review.review_time_desc

    @property
    def star_rating(self):
        return to_int_or_none(
            self._core_review.star_rating
        )



API_URL = 'https://www.tsd-aff.com/cgi-bin/xml_core.exe'
EXT_START_SESSION_URL = 'https://www.tsd-aff.com/cgi-bin/xml_start_session.exe'

DATE_SLUG_MAP = (
    {
        'slug': 'today',
        'auto_date': 'today_only',
    },
    {
        'slug': 'this-week',
        'auto_date': 'next_week',
    }
)

AUTO_DATE_RANGE = (
    'today_only', 'this_weekend', 'next_week',
    'this_month', 'next_month',
)

NO_TIME_DESCR = 'Select'

DEFAULT_CONCESSION_DESCR = 'Standard'

SMALL_IMAGE = 'square'
MEDIUM_IMAGE = 'landscape'
LARGE_IMAGE = 'triplet_one'
LARGE_IMAGE_TWO = 'triplet_two'
LARGE_IMAGE_THREE = 'triplet_three'
LARGE_IMAGE_FOUR = 'triplet_four'
LARGE_IMAGE_FIVE = 'triplet_five'
MARQUEE_IMAGE = 'marquee'
SEATING_IMAGE = 'seating_plan'
SUPPLIER_IMAGE = 'supplier'

REQUEST_MEDIA = [
    SMALL_IMAGE, MEDIUM_IMAGE, LARGE_IMAGE, LARGE_IMAGE_TWO,
    LARGE_IMAGE_THREE, LARGE_IMAGE_FOUR, LARGE_IMAGE_FIVE,
    MARQUEE_IMAGE, SEATING_IMAGE, SUPPLIER_IMAGE,
]

LARGE_IMAGE_LIST = [
    LARGE_IMAGE, LARGE_IMAGE_TWO, LARGE_IMAGE_THREE,
    LARGE_IMAGE_FOUR, LARGE_IMAGE_FIVE,
]

# timeout in seconds for API requests
API_REQUEST_TIMEOUT = 120



from operator import attrgetter
from copy import deepcopy

from base import InterfaceObject, Seat, SeatBlock, Currency, Commission
from pyticketswitch.util import (
    format_price_with_symbol, to_float_or_none, to_float_summed,
    to_int_or_none, resolve_boolean, day_mask_to_bool_list, yyyymmdd_to_date,
)


class TicketType(InterfaceObject):
    """Object that represents both a TSW ticket type and price band.

    The TSW ticket type and price band elements are combined into a single
    object here. This was done to simplify the interface and the ticket type
    and price band elements are closely related. It would be rare to require
    information from a ticket type but not the associated price bands.

    Operations and information associated with either element are available
    through this object. However, unlike the Performance and Event objects,
    this object's attributes will not be populated on-demand, so they may
    only have a value if the object has been retrieved through the Performance
    object.

    Args:
        ticket_type_id (string): Id of the TicketType
        settings (kwargs): See Core constructor.
    """

    def __init__(
        self,
        ticket_type_id,
        core_ticket_type=None,
        core_price_band=None,
        core_currency=None,
        **settings
    ):
        self.ticket_type_id = ticket_type_id
        self._core_ticket_type = core_ticket_type
        self._core_price_band = core_price_band
        self._core_currency = core_currency
        self.blanket_discount_only = None
        self.quantity_options = None
        self._example_seats = None
        self._possible_concessions = False
        self._available_seat_blocks = False

        super(TicketType, self).__init__(**settings)

    def _get_cache_key(self):
        return self.ticket_type_id

    def set_valid_quantities(self, valid_quantities):
        """For internal use, sets the valid ticket quantities."""
        self.quantity_options = [
            (v, v) for v in valid_quantities if int(v) <= int(
                self._core_price_band.number_available
            )
        ]

    def get_concessions(
        self, no_of_tickets, despatch_method=None, trolley=None,
        seat_block_id=None, seat_block_offset=None,
        include_user_commission=None,
    ):
        """Retrieves the Concession objects for this TicketType.

        Returns the list of Concession objects for each ticket requested,
        so the reponse will be a list of lists that contain Concession
        objects. Internally the 'discount_options' API method is being
        called.

        Args:
            no_of_tickets (int): Number of tickets.
            despatch_method (DespatchMethod): Optional, the DespatchMethod
                object to be used.
            trolley (Trolley): Optional, an existing Trolley that this
                order will be added to.
            seat_block_id (string): Optional, the ID of the SeatBlock to
                reserve seats from. Can only be used if returned in the
                'available_seat_blocks' attribute.
            seat_block_offset (int): Optional, the offset into the seat block,
                with 0 being the start of the block. Can only be specified if
                a seat_block is provided.
            include_user_commission (boolean): Optional, flag to indicate
                whether to request the user's commission information.

        Returns:
            list: A list for each ticket requested of Concession objects (
                so a list of lists of Concession objects).
        """

        crypto_block = self._get_crypto_block_for_object(
            method_name='availability_options',
            interface_object=self
        )

        no_of_tickets = str(int(no_of_tickets))

        if despatch_method is not None:
            despatch_id = despatch_method.despatch_id
        else:
            despatch_id = None

        if trolley is not None:
            trolley_id = trolley.trolley_id
        else:
            trolley_id = None

        if seat_block_offset is not None:
            seat_block_offset = str(int(seat_block_offset))

        resp_dict = self.get_core_api().discount_options(
            crypto_block=crypto_block,
            upfront_data_token=self.settings['upfront_data_token'],
            band_token=self.ticket_type_id,
            no_of_tickets=no_of_tickets,
            despatch_token=despatch_id,
            trolley_token=trolley_id,
            seat_block_token=seat_block_id,
            seat_block_offset=seat_block_offset,
            add_user_commission=include_user_commission,
        )

        self._set_crypto_block(
            crypto_block=resp_dict['crypto_block'],
            method_name='discount_options'
        )

        if 'blanket_discount_only' in resp_dict:

            ticket_concessions = []

            for discounts in resp_dict['discounts']:
                concessions = []
                for discount in discounts:
                    con = Concession(
                        concession_id=discount.discount_token,
                        core_discount=discount,
                        core_currency=resp_dict['currency'],
                        **self._internal_settings()
                    )
                    concessions.append(con)

                if concessions:
                    ticket_concessions.append(concessions)

                    self._set_crypto_for_objects(
                        crypto_block=resp_dict['crypto_block'],
                        method_name='discount_options',
                        interface_objects=concessions
                    )

            self.blanket_discount_only = resp_dict.get('blanket_discount_only')
            self.ticket_concessions = ticket_concessions

            return ticket_concessions

        else:
            # Backend doesn't support discount codes
            # crypto_block should be passed to create order
            return None

    @property
    def description(self):
        if self._core_price_band and self._core_price_band.price_band_desc:
            return u"{0} - {1}".format(
                self._core_ticket_type.ticket_type_desc,
                self._core_price_band.price_band_desc
            )

        return self._core_ticket_type.ticket_type_desc

    @property
    def code(self):
        return self._core_ticket_type.ticket_type_code

    @property
    def price_band_code(self):
        return self._core_price_band.price_band_code

    @property
    def price_without_surcharge(self):
        """Formatted string value of the price excluding surcharge with
        currency symbol."""

        return format_price_with_symbol(
            self._core_price_band.ticket_price,
            self._core_currency.currency_pre_symbol,
            self._core_currency.currency_post_symbol
        )

    @property
    def price_without_surcharge_float(self):
        """Float value of the price excluding surcharge."""
        return to_float_or_none(self._core_price_band.ticket_price)

    @property
    def surcharge(self):
        """Formatted string value of the surcharge with currency symbol."""

        return format_price_with_symbol(
            self._core_price_band.surcharge,
            self._core_currency.currency_pre_symbol,
            self._core_currency.currency_post_symbol
        )

    @property
    def surcharge_float(self):
        """Float value of the surcharge."""
        return to_float_or_none(self._core_price_band.surcharge)

    @property
    def price_combined_float(self):
        """Float value of the combined price."""
        if self._core_price_band.combined:
            return to_float_or_none(
                self._core_price_band.combined
            )
        else:
            return to_float_summed(
                self._core_price_band.ticket_price,
                self._core_price_band.surcharge
            )

    @property
    def price_combined(self):
        """Formatted string value of the combined price with currency symbol.
        """
        formatted_price = None

        combined_price = str(self.price_combined_float)

        formatted_price = format_price_with_symbol(
            combined_price,
            self._core_currency.currency_pre_symbol,
            self._core_currency.currency_post_symbol
        )

        return formatted_price

    @property
    def number_available(self):
        """Number of tickets that can be booked for this TicketType."""
        return to_int_or_none(
            self._core_price_band.number_available
        )

    @property
    def total_seats(self):
        """Number of available seats for this TicketType."""
        return to_int_or_none(
            self._core_price_band.raw_total_seats
        )

    @property
    def contiguous_seats(self):
        """Number of contiguous available seats for this TicketType."""
        return to_int_or_none(
            self._core_price_band.raw_contiguous_seats
        )

    @property
    def is_offer(self):
        """Boolean to indicate if the default concession for this ticket
        type is a special offer.
        """
        return resolve_boolean(self._core_price_band.is_offer)

    @property
    def int_percentage_saving(self):
        """Integer value of the percentage saving."""
        return to_int_or_none(
            self._core_price_band.percentage_saving
        )

    @property
    def percentage_saving(self):
        """Formatted string value of the percentage saving with a '%' symbol.
        """
        per_sav = None

        if self._core_price_band.percentage_saving:
            if self._core_price_band.percentage_saving == '0':

                per_sav = None
            else:

                per_sav = self._core_price_band.percentage_saving + '%'

        return per_sav

    @property
    def non_offer_combined_float(self):
        """Float value of the original combined price (i.e. if there was no
        offer).
        """
        if self._core_price_band.combined:
            return to_float_or_none(
                self._core_price_band.non_offer_combined
            )
        else:
            return to_float_summed(
                self._core_price_band.non_offer_ticket_price,
                self._core_price_band.non_offer_surcharge
            )

    @property
    def non_offer_combined(self):
        """Formatted string value of the original combined price with
        currency symbol (i.e. if there was no offer).
        """
        formatted_price = None

        combined_price = str(self.non_offer_combined_float)

        formatted_price = format_price_with_symbol(
            combined_price,
            self._core_currency.currency_pre_symbol,
            self._core_currency.currency_post_symbol
        )

        return formatted_price

    @property
    def non_offer_seatprice_float(self):
        """Float value of the original seatprice (i.e. if there was no
        offer).
        """
        return to_float_or_none(self._core_price_band.non_offer_seatprice)

    @property
    def non_offer_seatprice(self):
        """Formatted string value of the original seatprice with
        currency symbol (i.e. if there was no offer).
        """
        seatprice = str(self.non_offer_seatprice_float)

        formatted_price = format_price_with_symbol(
            seatprice,
            self._core_currency.currency_pre_symbol,
            self._core_currency.currency_post_symbol
        )

        return formatted_price

    @property
    def non_offer_surcharge_float(self):
        """Float value of the original surcharge (i.e. if there was no
        offer).
        """
        return to_float_or_none(self._core_price_band.non_offer_surcharge)

    @property
    def non_offer_surcharge(self):
        """Formatted string value of the original surcharge with
        currency symbol (i.e. if there was no offer).
        """
        surcharge = str(self.non_offer_surcharge_float)

        formatted_price = format_price_with_symbol(
            surcharge,
            self._core_currency.currency_pre_symbol,
            self._core_currency.currency_post_symbol
        )

        return formatted_price

    @property
    def example_seats(self):
        """List of Seat objects representing expected/example seats."""
        if self._example_seats is None:

            self._example_seats = []

            for seat in self._core_price_band.example_seats:
                self._example_seats.append(
                    Seat(core_seat=seat)
                )

            self._example_seats.sort(key=attrgetter(
                'column_sort_id', 'row_sort_id'
            ))

        return self._example_seats

    @property
    def has_example_seat_with_restricted_view(self):
        """Boolean to indicate if any of the example seats have restricted
        views.

        Returns True if any of the example seats are marked as having a
        restricted view, False if all of the example seats are marked as
        not having a restricted view and None if the information is not
        provided.
        """
        ret_val = None
        for s in self.example_seats:
            if s.is_restricted_view:
                return True
            elif s.is_restricted_view is False:
                ret_val = False

        return ret_val

    @property
    def unique_example_seat_text(self):
        """List of unique example seat text strings."""
        seat_text = []
        for s in self.example_seats:
            if s.seat_text and s.seat_text not in seat_text:
                seat_text.append(s.seat_text)

        return seat_text

    @property
    def example_seats_are_real(self):
        """Boolean to indicate if the example_seats are actual seat numbers.

        If True, the seats in the example_seats list will be the seats that
        are reserved.
        """
        return resolve_boolean(self._core_price_band.example_seats_are_real)

    @property
    def possible_concessions(self):
        """List of Concession objects that are expected for this TicketType.

        Only available if explicitly requested when getting availability.
        """

        if self._possible_concessions is False:

            if self._core_price_band.possible_discounts:
                self._possible_concessions = []

                for disc in self._core_price_band.possible_discounts:

                    self._possible_concessions.append(
                        Concession(
                            core_discount=disc,
                            core_currency=self._core_currency
                        )
                    )

            else:
                self._possible_concessions = None

        return self._possible_concessions

    @property
    def available_seat_blocks(self):
        """List of SeatBlock objects that can be selected from for this
        TicketType.

        Only available if requested at the availability stage with the
        include_available_seat_blocks flag.
        """
        if self._available_seat_blocks is False:

            if self._core_price_band.free_seat_blocks:
                self._available_seat_blocks = []

                for sb in self._core_price_band.free_seat_blocks:

                    self._available_seat_blocks.append(
                        SeatBlock(core_seat_block=sb)
                    )

            else:
                self._available_seat_blocks = None

        return self._available_seat_blocks

    @property
    def default_concession_code(self):
        """The concession code for the default concession of this TicketType"""
        return self._core_price_band.discount_code

    @property
    def user_commission(self):
        """Returns Commission object representing the user commission for the
        default concession on this TicketType.

        Only available if requested at the availability stage with the
        include_user_commission flag.
        """
        if self._core_price_band.user_commission:
            return Commission(
                core_commission=self._core_price_band.user_commission
            )
        return None

    @property
    def gross_commission(self):
        """Returns Commission object representing the gross commission for the
        default concession on this TicketType.

        Only available if requested at the availability stage with the
        include_user_commission flag and the gross commission is visible to the
        user.
        """
        if self._core_price_band.gross_commission:
            return Commission(
                core_commission=self._core_price_band.gross_commission
            )
        return None


class Concession(InterfaceObject):
    """Object that represents a TSW discount.

    Information relating to a TSW discount is accessible with this object.
    Was renamed to Concession as this was felt to be more appropriate in
    most cases.

    Can be constructed with the Id of the Concession. However, unlike the
    Performance and Event objects, this object's attributes will not be
    populated on-demand, so they may only have a value if the object has
    been retrieved through the TicketType object.

    Args:
        concession_id (string): Id of the Concession
        settings (kwargs): See Core constructor.
    """

    def __init__(
        self,
        concession_id=None,
        core_discount=None,
        core_currency=None,
        **settings
    ):
        self.concession_id = concession_id
        self._core_discount = core_discount
        self._core_currency = core_currency
        self._seats = None

        super(Concession, self).__init__(**settings)

    def _get_cache_key(self):
        return self.concession_id

    @property
    def description(self):
        discount_desc = self._core_discount.discount_desc

        if not discount_desc:
            discount_desc = self.settings['default_concession_descr']

        return discount_desc

    @property
    def code(self):
        """Code for internal use."""
        return self._core_discount.discount_code

    @property
    def ticket_price(self):
        """Formatted string value of the combined price with currency symbol.
        """
        return format_price_with_symbol(
            self._core_discount.combined,
            self._core_currency.currency_pre_symbol,
            self._core_currency.currency_post_symbol
        )

    @property
    def ticket_price_float(self):
        """Float value of the combined price."""
        return to_float_or_none(
            self._core_discount.combined
        )

    @property
    def no_of_tickets(self):
        return to_int_or_none(
            self._core_discount.no_of_tickets
        )

    @property
    def surcharge(self):
        """Formatted string value of the surcharge with currency symbol.
        """
        return format_price_with_symbol(
            self._core_discount.surcharge,
            self._core_currency.currency_pre_symbol,
            self._core_currency.currency_post_symbol
        )

    @property
    def surcharge_float(self):
        """Float value of the surcharge."""
        return to_float_or_none(self._core_discount.surcharge)

    @property
    def seatprice(self):
        """Formatted string value of the seat price (i.e. without surcharge)
        with currency symbol.
        """
        return format_price_with_symbol(
            self._core_discount.seatprice,
            self._core_currency.currency_pre_symbol,
            self._core_currency.currency_post_symbol
        )

    @property
    def seatprice_float(self):
        """Float value of the seat price (i.e. without surcharge)."""
        return to_float_or_none(self._core_discount.seatprice)

    @property
    def seats(self):
        """List of Seat objects for this Concession.

        This attribute will only have a value if the Concession was
        accessed through a Reservation object.
        """
        if self._seats is None:
            self._seats = []

            for seat in self._core_discount.seats:
                self._seats.append(
                    Seat(core_seat=seat)
                )

        return self._seats

    @property
    def seat_ids(self):
        """List of Seat Ids for this Concession.

        This attribute will only have a value if the Concession was
        accessed through a Reservation object.
        """
        return [s.seat_id for s in self.seats if s.seat_id]

    @property
    def has_restricted_view(self):
        """Boolean indcating if any seats have a restricted view.

        This attribute will only have a value if the Concession was
        accessed through a Reservation object.
        """
        restricted = False
        for s in self.seats:
            if s.is_restricted_view:
                restricted = True

        return restricted

    @property
    def unique_seat_text(self):
        """Information about this seat, should be displayed if present.

        This attribute will only have a value if the Concession was
        accessed through a Reservation object.
        """
        seat_text = []
        for s in self.seats:
            if s.seat_text and s.seat_text not in seat_text:
                seat_text.append(s.seat_text)

        return seat_text

    @property
    def number_available(self):
        """Number of tickets that can be booked for this Concession."""
        return to_int_or_none(
            self._core_discount.number_available
        )

    @property
    def total_seats(self):
        """Number of available seats for this Concession."""
        return to_int_or_none(
            self._core_discount.raw_total_seats
        )

    @property
    def contiguous_seats(self):
        """Number of contiguous available seats for this Concession."""
        return to_int_or_none(
            self._core_discount.raw_contiguous_seats
        )

    @property
    def user_commission(self):
        """Returns Commission object representing the user commission for this
        Concession.

        Only available if requested in get_concessions with the
        include_user_commission flag.
        """
        if self._core_discount.user_commission:
            return Commission(
                core_commission=self._core_discount.user_commission
            )
        return None

    @property
    def gross_commission(self):
        """Returns Commission object representing the gross commission for this
        Concession.

        Only available if requested in get_concessions with the
        include_user_commission flag and the gross commission is visible to the
        user.
        """
        if self._core_discount.gross_commission:
            return Commission(
                core_commission=self._core_discount.gross_commission
            )
        return None


class DespatchMethod(InterfaceObject):
    """Object that represents a TSW despatch method.

    Information relating to a TSW despatch method is available with
    this object. Can be constructed with the Id of the DespatchMethod
    (the other constructor arguments are for internal use).

    Args:
        despatch_id (string): Id of the DespatchMethod
        settings (kwargs): See Core constructor.
    """

    def __init__(
        self,
        despatch_id=None,
        core_despatch_method=None,
        core_currency=None,
        **settings
    ):
        self.despatch_id = despatch_id
        self._core_despatch_method = core_despatch_method
        self._core_currency = core_currency

        super(DespatchMethod, self).__init__(**settings)

    @property
    def description(self):
        return self._core_despatch_method.despatch_desc

    @property
    def code(self):
        """Code for internal use."""
        return self._core_despatch_method.despatch_code

    @property
    def cost(self):
        """Formatted string value of the cost with currency symbol."""
        return format_price_with_symbol(
            self._core_despatch_method.despatch_cost,
            self._core_currency.currency_pre_symbol,
            self._core_currency.currency_post_symbol
        )

    @property
    def cost_float(self):
        """Float value of the cost."""
        return to_float_or_none(
            self._core_despatch_method.despatch_cost
        )

    @property
    def currency(self):
        if self._core_currency:
            return Currency(
                core_currency=self._core_currency
            )

        return None


class AvailDetail(InterfaceObject):
    """Object that represents an individual availability detail for an event.

    An availability detail represents pricing information for a given offer,
    including ticket type, price band, which days of the week that price is
    available and for what date range.

    In fact, this AvailDetail object combines the TSW AvailDetail object with
    the related ticket type code+desc and price band code+desc since it is
    more convenient to represent it this way.

    This object does not make use of the TicketType object since we don't have
    access to the real TSW ticket type and therefore have no use for nearly
    all of its functionality.
    """

    def __init__(
        self,
        core_avail_detail,
        ticket_type_code=None,
        ticket_type_desc=None,
        price_band_code=None,
        price_band_desc=None,
        **settings
    ):

        self._core_avail_detail = core_avail_detail
        self._ticket_type_code = ticket_type_code
        self._ticket_type_desc = ticket_type_desc
        self._price_band_code = price_band_code
        self._price_band_desc = price_band_desc

        self._core_currency = core_avail_detail.currency

        self._available_from_date = None
        self._available_until_date = None
        self._weekdays_available = None

        super(AvailDetail, self).__init__(**settings)

    def is_same_ticket_and_price(self, other):
        """Compares avail details with another. Returns True if
        ticket_type_desc, seatprice, surcharge and int_percentage_saving
        are the same.
        """

        result = False

        if isinstance(other, self.__class__):
            if (
                self.ticket_type_desc == other.ticket_type_desc and
                self.seatprice == other.seatprice and
                self.surcharge == other.surcharge and
                self.int_percentage_saving == other.int_percentage_saving
            ):
                result = True

        return result

    def combine(self, other):
        """Combines avail detail with another. Does no checking to make sure
        avail details are actually equal (i.e. is_same_ticket_and_price should
        be called prior to this). Combining is done by 'OR'ing the
        weekdays_available and taking the widest available date range.
        """

        new_ad = deepcopy(self)

        # Do a union of weekdays available and date range
        if other.available_from_date < self.available_from_date:
            new_ad._available_from_date = other.available_from_date
        if other.available_until_date > self.available_until_date:
            new_ad._available_until_date = other.available_until_date

        new_ad._weekdays_available = [
            a | b for a, b in zip(
                self.weekdays_available, other.weekdays_available
            )
        ]

        return new_ad

    @property
    def ticket_type_code(self):
        return self._ticket_type_code or None

    @property
    def ticket_type_desc(self):
        return self._ticket_type_desc or None

    @property
    def price_band_code(self):
        return self._price_band_code or None

    @property
    def price_band_desc(self):
        return self._price_band_desc or None

    @property
    def available_from_date(self):

        if self._available_from_date is None:

            avail_dates = self._core_avail_detail.available_dates
            if avail_dates:
                from_date = avail_dates.get('first_yyyymmdd')
                if from_date:
                    self._available_from_date = yyyymmdd_to_date(from_date)

        return self._available_from_date

    @property
    def available_until_date(self):

        if self._available_until_date is None:

            avail_dates = self._core_avail_detail.available_dates
            if avail_dates:
                until_date = avail_dates.get('last_yyyymmdd')
                if until_date:
                    self._available_until_date = yyyymmdd_to_date(until_date)

        return self._available_until_date

    @property
    def weekdays_available(self):
        """Returns a list of booleans of length 7, one for each day of the
        week (starting with Sunday). Each boolean represents whether this
        ticket is available for that day of the week.
        """

        if self._weekdays_available is None:

            # This data comes back from TSW in the form of a bitmask, so we
            # first need to convert this into a binary representation
            self._weekdays_available = day_mask_to_bool_list(
                self._core_avail_detail.day_mask
            )

        return self._weekdays_available

    @property
    def weekdays_available_iso(self):
        """Same as weekdays_available but has week starting on Monday
        """

        weekdays_avail = self.weekdays_available
        # Shift list
        weekdays_avail_iso = weekdays_avail[1:] + weekdays_avail[:1]
        return weekdays_avail_iso

    @property
    def seatprice(self):
        """Formatted string value of the seatprice with currency symbol."""

        return format_price_with_symbol(
            self._core_avail_detail.seatprice,
            self._core_currency.currency_pre_symbol,
            self._core_currency.currency_post_symbol
        )

    @property
    def seatprice_float(self):
        """Float value of the seatprice."""

        return to_float_or_none(self._core_avail_detail.seatprice)

    @property
    def surcharge(self):
        """Formatted string value of the surcharge with currency symbol."""

        return format_price_with_symbol(
            self._core_avail_detail.surcharge,
            self._core_currency.currency_pre_symbol,
            self._core_currency.currency_post_symbol
        )

    @property
    def surcharge_float(self):
        """Float value of the surcharge"""

        return to_float_or_none(self._core_avail_detail.surcharge)

    @property
    def has_no_booking_fee(self):
        """Returns true if the surcharge is zero, false otherwise."""

        surcharge = to_float_or_none(self._core_avail_detail.surcharge)
        return (not bool(surcharge))

    @property
    def price_combined(self):
        """Formatted string value of the combined seatprice + surcharge with
        currency symbol.
        """

        combined_price = str(self.price_combined_float)

        return format_price_with_symbol(
            combined_price,
            self._core_currency.currency_pre_symbol,
            self._core_currency.currency_post_symbol
        )

    @property
    def price_combined_float(self):
        """Float value of the combined seatprice + surcharge
        """

        return to_float_summed(
            self._core_avail_detail.seatprice,
            self._core_avail_detail.surcharge
        )

    @property
    def non_offer_seatprice(self):
        """Formatted string value of the full_seatprice with currency symbol.
        Returns None if doesn't exist.
        """

        if self._core_avail_detail.full_seatprice:
            return format_price_with_symbol(
                self._core_avail_detail.full_seatprice,
                self._core_currency.currency_pre_symbol,
                self._core_currency.currency_post_symbol
            )

        return None

    @property
    def non_offer_seatprice_float(self):
        """Float value of the full_seatprice.
        Returns None if doesn't exist.
        """

        if self._core_avail_detail.full_seatprice:
            return to_float_or_none(self._core_avail_detail.full_seatprice)

        return None

    @property
    def non_offer_surcharge(self):
        """Formatted string value of the full_surcharge with currency symbol.
        Returns None if doesn't exist.
        """

        if self._core_avail_detail.full_surcharge:
            return format_price_with_symbol(
                self._core_avail_detail.full_surcharge,
                self._core_currency.currency_pre_symbol,
                self._core_currency.currency_post_symbol
            )

        return None

    @property
    def non_offer_surcharge_float(self):
        """Float value of the full_surcharge.
        Returns None if doesn't exist.
        """

        if self._core_avail_detail.full_surcharge:
            return to_float_or_none(self._core_avail_detail.full_surcharge)

        return None

    @property
    def non_offer_combined(self):
        """Formatted string value of the combined full_seatprice +
        full_surcharge with currency symbol. Returns None if doesn't exist.
        """

        if (self._core_avail_detail.full_seatprice and
                self._core_avail_detail.full_surcharge):

            combined_price = str(
                float(self._core_avail_detail.full_seatprice) +
                float(self._core_avail_detail.full_surcharge)
            )
            return format_price_with_symbol(
                combined_price,
                self._core_currency.currency_pre_symbol,
                self._core_currency.currency_post_symbol
            )

        return None

    @property
    def non_offer_combined_float(self):
        """Float value of the combined full_seatprice + full_surcharge.
        Returns None if doesn't exist.
        """

        if (self._core_avail_detail.full_seatprice and
                self._core_avail_detail.full_surcharge):

            return to_float_summed(
                self._core_avail_detail.full_seatprice,
                self._core_avail_detail.full_surcharge
            )

        return None

    @property
    def absolute_saving(self):
        """Formatted string value of the absolute_saving with currency symbol.
        Returns None if doesn't exist.
        """

        if self._core_avail_detail.absolute_saving:
            return format_price_with_symbol(
                self._core_avail_detail.absolute_saving,
                self._core_currency.currency_pre_symbol,
                self._core_currency.currency_post_symbol
            )

        return None

    @property
    def absolute_saving_float(self):
        """Float value of the absolute_saving.
        Returns None if doesn't exist.
        """

        if self._core_avail_detail.absolute_saving:
            return to_float_or_none(self._core_avail_detail.absolute_saving)

        return None

    @property
    def int_percentage_saving(self):
        """Integer value of the percentage saving."""
        return to_int_or_none(
            self._core_avail_detail.percentage_saving
        )

    @property
    def percentage_saving(self):
        """Formatted string value of the percentage_saving with currency symbol.
        Returns None if doesn't exist.
        """

        if self._core_avail_detail.percentage_saving:
            return u'{0}%'.format(self._core_avail_detail.percentage_saving)

        return None
